/**
 * Copyright (C) 2009-2012 enStratus Networks Inc
 *
 * ====================================================================
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ====================================================================
 */

package org.dasein.cloud.opsource.compute;


import java.text.DateFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Locale;


import javax.annotation.Nonnull;

import org.apache.log4j.Logger;
import org.dasein.cloud.CloudException;
import org.dasein.cloud.InternalException;
import org.dasein.cloud.OperationNotSupportedException;
import org.dasein.cloud.Tag;

import org.dasein.cloud.compute.Architecture;
import org.dasein.cloud.compute.MachineImage;
import org.dasein.cloud.compute.Platform;
import org.dasein.cloud.compute.VirtualMachine;
import org.dasein.cloud.compute.VirtualMachineProduct;
import org.dasein.cloud.compute.VirtualMachineSupport;
import org.dasein.cloud.compute.VmState;
import org.dasein.cloud.compute.VmStatistics;

import org.dasein.cloud.dc.Region;
import org.dasein.cloud.identity.ServiceAction;
import org.dasein.cloud.opsource.OpSource;
import org.dasein.cloud.opsource.OpSourceMethod;
import org.dasein.cloud.opsource.Param;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

public class VirtualMachines implements VirtualMachineSupport {
    static public final Logger logger = Logger.getLogger(VirtualMachines.class);
    
   // static private final String ADD_STORAGE_TO_VIRTUAL_MACHINE    = "addLocalStorage";
    //static private final String DEPLOY_VIRTUAL_MACHINE  = "deployVirtualMachine";
    static private final String DESTROY_VIRTUAL_MACHINE = "delete";
    static private final String CLEAN_VIRTUAL_MACHINE = "clean";
    static private final String REBOOT_VIRTUAL_MACHINE  = "reboot";
   // static private final String RESET_VIRTUAL_MACHINE   = "reset";
    static private final String START_VIRTUAL_MACHINE   = "start";
    static private final String PAUSE_VIRTUAL_MACHINE 	= "shutdown";
   // static private final String STOP_VIRTUAL_MACHINE    = "poweroff";    

    
	//Node tag name
	
	static private final String Deployed_Server_Tag = "Server";
	static private final String Pending_Deployed_Server_Tag = "PendingDeployServer";
	
   // static private final String UPDATE_VMWARE_TOOL   = "updateVMwareTools";
  
    
	int attemptForOperation = 30;
	long waitTimeToAttempt = 30000L;
    
    private OpSource provider;
    
    public VirtualMachines(OpSource provider) {
        this.provider = provider;
    }
    
    
    public boolean attachDisk(String serverId, int sizeInGb) throws InternalException, CloudException {
    	HashMap<Integer, Param>  parameters = new HashMap<Integer, Param>();
        Param param = new Param(OpSource.SERVER_BASE_PATH, null);
     	parameters.put(0, param);
     	
     	param = new Param(serverId, null);
     	parameters.put(1, param);   
    	
     	param = new Param("amount", String.valueOf(sizeInGb));
     	parameters.put(2, param); 
     	
     	
     	OpSourceMethod method = new OpSourceMethod(provider, 
     			provider.buildUrl("addLocalStorage",true, parameters),
     			provider.getBasicRequestParameters(OpSource.Content_Type_Value_Single_Para, "GET", null));
       	
     	Document doc = method.invoke();
     	
     	return method.parseRequestResult("Attaching disk", doc , "ns9:result","ns9:resultDetail");
    }

    
    @Override
    public void boot(String serverId) throws InternalException, CloudException {
       	//this.modify(serverId, 1, 1024);
       	start(serverId);
    }
    
    private boolean start(String serverId) throws InternalException, CloudException {
       	
        HashMap<Integer, Param>  parameters = new HashMap<Integer, Param>();
        Param param = new Param(OpSource.SERVER_BASE_PATH, null);
    	parameters.put(0, param);
    	param = new Param(serverId, null);
    	parameters.put(1, param); 
    	
    	OpSourceMethod method = new OpSourceMethod(provider, 
    			provider.buildUrl(START_VIRTUAL_MACHINE,true, parameters),
    			provider.getBasicRequestParameters(OpSource.Content_Type_Value_Single_Para, "GET",null));
      	return method.parseRequestResult("Booting vm",method.invoke(), "ns9:result", "ns9:resultDetail");
    }
    
    private boolean cleanFailedVM(String serverId) throws InternalException, CloudException {
     	
    	HashMap<Integer, Param>  parameters = new HashMap<Integer, Param>();
        Param param = new Param(OpSource.SERVER_BASE_PATH, null);
      	parameters.put(0, param);
      	param = new Param(serverId, null);
      	parameters.put(1, param); 
      	
      	OpSourceMethod method = new OpSourceMethod(provider, 
      			provider.buildUrl(CLEAN_VIRTUAL_MACHINE,true, parameters),
      			provider.getBasicRequestParameters(OpSource.Content_Type_Value_Single_Para, "GET",null));
        	return method.parseRequestResult("Clean failed vm",method.invoke(),"ns9:result", "ns9:resultDetail");
    }
    
    
    @Override
    public VirtualMachine clone(String serverId, String intoDcId, String name, String description, boolean powerOn, String ... firewallIds) throws InternalException, CloudException {
        throw new OperationNotSupportedException("Instances cannot be cloned.");
    }
    

    @Override
    public void disableAnalytics(String vmId) throws InternalException, CloudException {
        // NO-OP        
    }

    @Override
    public void enableAnalytics(String vmId) throws InternalException, CloudException {
        // NO-OP
    }
    
    @Override
    public String getConsoleOutput(String serverId) throws InternalException, CloudException {
        return "";
    }
    
    @Override 
    public VirtualMachineProduct getProduct(String productId) throws InternalException, CloudException {
        for( Architecture architecture : Architecture.values() ) {
            for( VirtualMachineProduct product : listProducts(architecture) ) {
                if( product.getProductId().equals(productId) ) {
                    return product;
                }
            }
        }
        if( logger.isDebugEnabled() ) {
            logger.debug("Unknown product ID for cloud.com: " + productId);
        }
        return null;
    }
    
    @Override
    public String getProviderTermForServer(Locale locale) {
        return "Server";
    }

    @Override
    public VirtualMachine getVirtualMachine(String serverId) throws InternalException, CloudException {
        HashMap<Integer, Param>  parameters = new HashMap<Integer, Param>();
        Param param = new Param(OpSource.SERVER_BASE_PATH, null);
    	parameters.put(0, param);
    	
    	param = new Param(serverId, null);
    	parameters.put(1, param);    
    	
    	OpSourceMethod method = new OpSourceMethod(provider, 
    			provider.buildUrl(null,true, parameters),
    			provider.getBasicRequestParameters(OpSource.Content_Type_Value_Single_Para, "GET",null));
    	
    	Document doc = method.invoke();
    	
    	
    	NodeList  matches = doc.getElementsByTagName("Server");
    	if(matches != null){
    		return toVirtualMachine(matches.item(0), false);      	
    	}    	
    	return null;    	
    }
       
    public VirtualMachine getVirtualMachineByName(String name) throws InternalException, CloudException {
    	ArrayList<VirtualMachine> list = (ArrayList<VirtualMachine>) listePendingServers();
    	for(VirtualMachine vm : list ){
    		if(vm.getName().equals(name)){
    			return vm;  			      			
    		}    		
    	}
     	list = (ArrayList<VirtualMachine>) listDeployServers();
    	for(VirtualMachine vm : list ){
    		if(vm.getName().equals(name)){
    			return vm;  			      			
    		}    		
    	}
    	return null;
    }
    
    
    @Override
    public VmStatistics getVMStatistics(String serverId, long startTimestamp, long endTimestamp) throws InternalException, CloudException {
        return new VmStatistics();
    }
    
    @Override
    public Iterable<VmStatistics> getVMStatisticsForPeriod(String arg0, long arg1, long arg2) throws InternalException, CloudException {
        return Collections.emptyList();
    }
    
    @Override
    public boolean isSubscribed() throws CloudException, InternalException {
    	return true;
    }

    @Override
    public VirtualMachine launch(String imageId, VirtualMachineProduct product, String inZoneId, String name, String description, String usingKey, String withVlanId, boolean withMonitoring, boolean asSandbox, String... protectedByFirewalls) throws InternalException, CloudException {
        return launch(imageId, product, inZoneId, name, description, usingKey, withVlanId, withMonitoring, asSandbox, protectedByFirewalls, new Tag[0]);
    
    }
    
    public VirtualMachine launch(String imageId, VirtualMachineProduct product, String inZoneId, String name, String description, String usingKey, String withVlanId, boolean withMonitoring, boolean asSandbox, String[] protectedByFirewalls, Tag ... tags) throws InternalException, CloudException {
    	
    	//First step get the target image
	  	ServerImage imageSupport = new ServerImage(provider) ;
    	
	  	MachineImage origImage = imageSupport.getOpSourceImage(imageId);
    	
    	if(origImage == null){
    		throw new CloudException("No such image to launch VM");
    	}
		
    	int targetCPU = product.getCpuCount();
		int targetMemory = product.getRamInMb();		
		int targetDisk = product.getDiskSizeInGb();
        	
    	int currentCPU = 0;
		int currentMemory =0; 
		int currentDisk = 10;
		
		if(origImage.getTag("cpuCount") != null){
			currentCPU = Integer.valueOf((String) origImage.getTag("cpuCount"));
		}
		
		if(origImage.getTag("memory") != null){
			currentMemory = Integer.valueOf((String) origImage.getTag("memory"));
		}
		
		if( targetDisk == 0 && currentCPU == targetCPU && currentMemory == targetMemory ){
			boolean isDeployed = this.deploy(origImage.getProviderMachineImageId(), inZoneId, name, description, withVlanId, null, "true");
			if(isDeployed){
				return getVirtualMachineByName(name);
			}else{
				throw new CloudException("Fail to launch the server");
			}    
			
		}else if(targetDisk == 0 && ((targetCPU == 1 && targetMemory == 2048) || (targetCPU == 2 && targetMemory == 4096) || (targetCPU == 4 && targetMemory == 6144))){
			// It is Opsource OS, then get the target image with the same cpu and memory
			MachineImage targetTmage = imageSupport.searchImage(origImage.getPlatform(), origImage.getArchitecture(), product.getCpuCount(), product.getRamInMb());
    		if(targetTmage != null){
    			boolean isDeployed = this.deploy(targetTmage.getProviderMachineImageId(), inZoneId, name, description, withVlanId, null, "true");
    			
    			if(isDeployed){
    				return getVirtualMachineByName(name);
    			}else{
    				throw new CloudException("Fail to launch the server");
    			}    
    		}        	
		}		
		
		//There is target image with the CPU and memory required, then need to modify the server after deploying
    	
    	//Second step deploy VM
    	boolean isDeployed = this.deploy(imageId, inZoneId, name, description, withVlanId, null, "false");
  
    	long starttime = System.currentTimeMillis();
    	
    	//Third Step Modify VM
    	if(!isDeployed){   
    		throw new CloudException("Fail to deploy VM");
    	} 
    	 	
    	VirtualMachine server = getVirtualMachineByName(name);
		//update the hardware (CPU, memory configuration);
    	
    	if(server == null){    		
    		throw new CloudException("Server failed to deployed without explaination");
    	}
    	
    	if(currentCPU != targetCPU || currentMemory != targetMemory ){			
    	   	        	
        	int localAttemptToUpdateVM = attemptForOperation;
    		
			//Modify server to target cpu and memory				
	   		while (localAttemptToUpdateVM >0){
				try {
					//VM has finished deployment before continuing, therefore wait 15s 
					server = getVirtualMachineByName(name);
					if(server == null){    		
			    		throw new CloudException("Server failed to launch during modifying the CPU and Memory !!!");
			    	}
					
					currentCPU = Integer.valueOf((String) server.getTag("cpuCount"));
					
					currentMemory = Integer.valueOf((String) server.getTag("memory"));
					
					if(currentCPU != targetCPU || currentMemory != targetMemory ){		
						logger.info("Begin to modify ->" + localAttemptToUpdateVM );
						boolean isModify = modify(server.getProviderVirtualMachineId(), targetCPU, targetMemory);
						
						if(isModify){
							
							localAttemptToUpdateVM = 0;
						}
					}else{
						
						localAttemptToUpdateVM = 0;
					}
				
				
				} catch (InternalException e) {
					//Throwable?
				} catch (CloudException e) {
					try{					
						Thread.sleep(waitTimeToAttempt);
						
						localAttemptToUpdateVM--;
					
					}catch(InterruptedException e1){ 
						throw new InternalException("Failed while trying to wait 30s to update the server with Id ");
					}   				
				} 		
			}
			long end = System.currentTimeMillis();
		
			logger.info("Total deploy time -> " + ((end-starttime)/1000));
	    	    		
    	}

		
		
		//Third Step: attach the disk
    	
    	if( targetDisk != currentDisk){
    		
        	//Update usually take another 6 mins
    		
    		starttime = System.currentTimeMillis();
    		
    		int localAttemptToAttachDisk = attemptForOperation;
    	
       		while (localAttemptToAttachDisk >0){
    			try {		
    				//Begin to attach the VM
    				logger.info("Begin to attach the server " + localAttemptToAttachDisk);
    				
    				server = getVirtualMachineByName(name);
    				
    				if(server == null){    		
    		    		throw new CloudException("Server failed to launch while attaching disk !!!");
    		    	}
    				    		
    				if(server.getProduct() != null && (server.getProduct().getDiskSizeInGb() == targetDisk) ){
    					
    					localAttemptToAttachDisk = 0;
    				}
    				
    				boolean isAttached = attachDisk(server.getProviderVirtualMachineId(), targetDisk);
    				
    				if(isAttached){
    					
    					localAttemptToAttachDisk = 0;
    				}    				
    				long end = System.currentTimeMillis();				
    				logger.info("Total attach time -> " + ((end-starttime)/1000));
    								
    			} catch (InternalException e) {
    				
    			} catch (CloudException e) {				
    				try{	
    					
    					Thread.sleep(waitTimeToAttempt);    					
    					localAttemptToAttachDisk--;
    					
    				}catch (InterruptedException e1) {
    					
    					throw new InternalException("Failed while trying to wait 30s to attaching disk to the server");
    				}		 	
    			} 	
    		}	
    		
    	}

		// Fourth Step: boot the server		
    	//Update usually take another 10 mins, wait 5 minutes first    	
		starttime = System.currentTimeMillis();
		
		int localAttemptToBootVM = attemptForOperation;
	
   		while (localAttemptToBootVM >0){
			try {		
				//Begin to start the VM
				
				server = getVirtualMachineByName(name);
				
				if(server == null){
					
		    		throw new CloudException("Server failed to launch while attaching disk !!!");
		    	}
				
				if(server.getCurrentState().equals(VmState.RUNNING)){    		
		    		//Already started					
					return server;
		    	}
								
				start(server.getProviderVirtualMachineId());
				
				long end = System.currentTimeMillis();				
				
				logger.info("Total boot time -> " + ((end-starttime)/1000));
				
				return server;
				
			} catch (InternalException e) {
				
			} catch (CloudException e) {				
				try{				
					
					Thread.sleep(waitTimeToAttempt);
					
					localAttemptToBootVM--;	
					
				}catch (InterruptedException e1) {
					throw new InternalException("Failed while trying to wait 30s to update the server with Id ");
				}		 	
			} 	
		}		
    	return null;
    }
    
    /**
     * 
     * @param imageId
     * @param inZoneId
     * @param name
     * @param description
     * @param withVlanId
     * @param adminPassword
     * @param isStart: true or false
     */    
    
    private boolean deploy(@Nonnull String imageId, String inZoneId, String name, String description, String withVlanId, String adminPassword, String isStart) throws InternalException, CloudException {
    	inZoneId = translateZone(inZoneId);
       
        //Create post body
        Document doc = provider.createDoc();
        Element server = doc.createElementNS("http://oec.api.opsource.net/schemas/server", "Server");
        
        Element nameElmt = doc.createElement("name");
        nameElmt.setTextContent(name);
       
        Element descriptionElmt = doc.createElement("description");
        descriptionElmt.setTextContent(description);
        
        if(withVlanId == null){
        	withVlanId = provider.getDefaultVlanId();
        }       
        
        Element vlanResourcePath = doc.createElement("vlanResourcePath");
        vlanResourcePath.setTextContent(provider.getVlanResourcePathFromVlanId(withVlanId));
    
        Element imageResourcePath = doc.createElement("imageResourcePath");
        imageResourcePath.setTextContent(provider.getImageResourcePathFromImaged(imageId));
        
        if(adminPassword == null){       	
        	adminPassword = provider.getDefaultAdminPasswordForVM();        	
        }else{
        	if(adminPassword.length() < 8){
        		throw new InternalException("Password require a minimum of 8 characters!!!");
        	}
        }
        
        Element administratorPassword = doc.createElement("administratorPassword");
        administratorPassword.setTextContent(adminPassword); 
        
        Element isStartElmt = doc.createElement("isStarted");
        
        isStartElmt.setTextContent(isStart); 
        
        doc.appendChild(server);        
        server.appendChild(nameElmt);
        server.appendChild(descriptionElmt);
        server.appendChild(vlanResourcePath);
        server.appendChild(imageResourcePath);
        server.appendChild(administratorPassword);
        
        server.appendChild(isStartElmt);
        
                
        HashMap<Integer, Param>  parameters = new HashMap<Integer, Param>();
        
        Param param = new Param(OpSource.SERVER_BASE_PATH, null);
    	parameters.put(0, param);
  
    	OpSourceMethod method = new OpSourceMethod(provider,
    			provider.buildUrl(null,true, parameters),
    			provider.getBasicRequestParameters(OpSource.Content_Type_Value_Single_Para, "POST", provider.convertDomToString(doc)));
    	return method.parseRequestResult("Deploying server",method.invoke(), "ns9:result", "ns9:resultDetail");
     }
    

    @Override
    public Iterable<String> listFirewalls(String vmId) throws InternalException, CloudException {
    	//Not support firewall
    	return Collections.emptyList();
    }
    
    @Override
    public Iterable<VirtualMachineProduct> listProducts(Architecture architecture) throws InternalException, CloudException {
        List<VirtualMachineProduct> products = new ArrayList<VirtualMachineProduct>();
      
        VirtualMachineProduct product;        
        //OpSource enables any combination of CPU (1 -8 ) and RAM (1 - 64G)
        int diskSizeInGb = 0;
        for( int disk = 0 ; disk < 6; disk ++ ){
        	diskSizeInGb = disk * 50;
        	
        	
            for(int cpuNum =1;cpuNum <= 8;cpuNum ++){
            	/**
        		 * Default cpuNum = 1, 2, max ram = 8
        		 * cpuNum = 3, 4, min ram 4, max ram = 32
        		 * cpuNum = 1, 2, max ram = 8
        		 */
            	int ramInGB = 1*cpuNum;
            	if(cpuNum <=2){
            		ramInGB = 1;
            	}
            	while(ramInGB <= 4*cpuNum ){
            		product = new VirtualMachineProduct();
            		product.setProductId(cpuNum+ ":" +ramInGB + ":" + disk);
            		product.setName(" (" + cpuNum + " CPU/" + ramInGB + " Gb RAM/" + diskSizeInGb + " Gb Disk)");
            		product.setDescription(" (" + cpuNum + " CPU/" + ramInGB + " Gb RAM/" + diskSizeInGb + " Gb Disk)");
            		product.setRamInMb(ramInGB*1024);
            		product.setCpuCount(cpuNum);
            		product.setDiskSizeInGb(diskSizeInGb);                
            		products.add(product);
            		
            		if(cpuNum <=2){
            			ramInGB = ramInGB + 1;
            		}else{
            			ramInGB = ramInGB + ramInGB;
            		}
            	}        	
            }       
        	
        }

  
        return products;
    }
    
    @Override
    public Iterable<VirtualMachine> listVirtualMachines() throws InternalException, CloudException {
    	ArrayList<VirtualMachine> allList = new ArrayList<VirtualMachine>();
    	ArrayList<VirtualMachine> list = (ArrayList<VirtualMachine>) listePendingServers();
    	if(list != null){
    		allList.addAll(list);
    	}
    	
    	list = (ArrayList<VirtualMachine>) listDeployServers();
    	if(list != null){
    		allList.addAll(list);
    	}
        return allList;
    }
    
    
    private Iterable<VirtualMachine> listDeployServers() throws InternalException, CloudException {
    	ArrayList<VirtualMachine> list = new ArrayList<VirtualMachine>();
         
    	//Get deploy Server
        HashMap<Integer, Param> parameters = new HashMap<Integer, Param>();
    	Param param = new Param(OpSource.SERVER_BASE_PATH, null);
    	parameters.put(0, param);    
    	
    	OpSourceMethod method = new OpSourceMethod(provider, 
    			provider.buildUrl(null, true, parameters),
    			provider.getBasicRequestParameters(OpSource.Content_Type_Value_Single_Para, "GET",null));
    	
    	
    	Document doc = method.invoke();
        
    	NodeList matches = doc.getElementsByTagName(Deployed_Server_Tag);
        
    	if(matches != null){
            for( int i=0; i<matches.getLength(); i++ ) {
                Node node = matches.item(i);            
                VirtualMachine vm = this.toVirtualMachine(node, false);
                if( vm != null ) {
                	list.add(vm);
                }
            }
        }         
        return list;
    }
    
   private Iterable<VirtualMachine> listePendingServers() throws InternalException, CloudException {
    	ArrayList<VirtualMachine> list = new ArrayList<VirtualMachine>();
        
    	//Get pending deploy Image
        HashMap<Integer, Param> parameters = new HashMap<Integer, Param>();
    	Param param = new Param(OpSource.SERVER_BASE_PATH, null);
    	parameters.put(0, param);
    	param = new Param("pendingDeploy", null);
    	parameters.put(1, param);
    
    	OpSourceMethod method = new OpSourceMethod(provider, 
    			provider.buildUrl(null, true, parameters),
    			provider.getBasicRequestParameters(OpSource.Content_Type_Value_Single_Para, "GET",null));
    	
    	Document doc = method.invoke();
    	
        NodeList matches = doc.getElementsByTagName(Pending_Deployed_Server_Tag);
        if(matches != null){
            for( int i=0; i<matches.getLength(); i++ ) {
                Node node = matches.item(i);            
                VirtualMachine vm = this.toVirtualMachine(node, true);
                if( vm != null ) {
                	list.add(vm);
                }
            }
        }         
        return list;
    } 
    

    @Override
    public @Nonnull String[] mapServiceAction(@Nonnull ServiceAction action) {
        return new String[0];
    }
    
    //Modify VM with the cpu and memory
    private boolean modify(String serverId, int cpuCount, int memoryInMb ) throws InternalException, CloudException {
       	
        HashMap<Integer, Param>  parameters = new HashMap<Integer, Param>();
        Param param = new Param(OpSource.SERVER_BASE_PATH, null);
    	parameters.put(0, param);
    	param = new Param(serverId, null);
    	parameters.put(1, param);
    	
        //Create post body
       	String requestBody = "cpuCount=";
       	requestBody += cpuCount;
       	requestBody += "&memory=" + memoryInMb;
       	   	
    	OpSourceMethod method = new OpSourceMethod(provider, 
    			provider.buildUrl(null,true, parameters),
    			provider.getBasicRequestParameters(OpSource.Content_Type_Value_Modify, "POST", requestBody));
      	return method.parseRequestResult("Modify vm",method.invoke(), "ns9:result", "ns9:resultDetail");
    }  

    @Override
    public void pause(String serverId) throws InternalException, CloudException {
    	 HashMap<Integer, Param>  parameters = new HashMap<Integer, Param>();
         Param param = new Param(OpSource.SERVER_BASE_PATH, null);
     	parameters.put(0, param);
     	param = new Param(serverId, null);
     	parameters.put(1, param); 
     	
     	//Gracefully power off
     	OpSourceMethod method = new OpSourceMethod(provider, 
     			provider.buildUrl(PAUSE_VIRTUAL_MACHINE,true, parameters),
     			provider.getBasicRequestParameters(OpSource.Content_Type_Value_Single_Para, "GET",null));
       	method.parseRequestResult("Pauseing vm",method.invoke(),"ns9:result","ns9:resultDetail");      
    }
 
    
    @Override
    public void reboot(String serverId) throws CloudException, InternalException {
       	
        HashMap<Integer, Param>  parameters = new HashMap<Integer, Param>();
        Param param = new Param(OpSource.SERVER_BASE_PATH, null);
    	parameters.put(0, param);
    	param = new Param(serverId, null);
    	parameters.put(1, param); 
    	
    	OpSourceMethod method = new OpSourceMethod(provider, 
    			provider.buildUrl(REBOOT_VIRTUAL_MACHINE,true, parameters),
    			provider.getBasicRequestParameters(OpSource.Content_Type_Value_Single_Para, "GET",null));
    	method.parseRequestResult("Rebooting vm",method.invoke(),"ns9:result","ns9:resultDetail");
    }

    @Override
    public boolean supportsAnalytics() throws CloudException, InternalException {
        return false;
    }
    
    @Override
    public void terminate(String serverId) throws InternalException, CloudException {
    	VirtualMachine server = getVirtualMachine(serverId);
    	
    	if(server == null){
    		throw new CloudException("No such VM");    		
    	}
    	
    	// Release public IP first
    	if(server.getPublicIpAddresses() != null){
    		for(String addressId : server.getPublicIpAddresses()){
    			provider.getNetworkServices().getIpAddressSupport().releaseFromServer(addressId);
    		}   		
    	}
        		
		//Pause the vm first
		
		int localAttemptToPauseVM = attemptForOperation;
		
		while (localAttemptToPauseVM >0){
			try { 
				//If it is pending, means it is in deployment process, need around 6 mins
				server = getVirtualMachine(serverId);
				
				if(server == null){
					//VM already killed
					return;    					
				}    				
				
				if(server.getCurrentState().equals(VmState.RUNNING)){
					
					pause(serverId);
					
				}else{
					
					localAttemptToPauseVM = 0;
				}
				
			} catch (InternalException e) {
				
			} catch (CloudException e){
				
				try {
					
					Thread.sleep(waitTimeToAttempt);
					
					localAttemptToPauseVM--;
					
				} catch (InterruptedException e1) {
					
	    			throw new InternalException ("Fail when waiting");		
				}    				
			}		
    		    	
    		int localAttemptToTerminateVM = attemptForOperation;
    	
       		while (localAttemptToTerminateVM >0){
    			try {    				
    				//Begin to Kill the VM
    				
    				server = getVirtualMachine(serverId);    				
    					
    				if(server == null){
    					//VM already killed
    					return;    					
    				}
    				
    				// In case the server is up from the booting stage
    				
    				if(server.getCurrentState().equals(VmState.RUNNING)){
    					
    					pause(serverId);
    					
    					localAttemptToTerminateVM = attemptForOperation; 
    					
    				}else{
    					
    					logger.info("Begin to kill -> " + localAttemptToTerminateVM);
    					
        				String  resultCode = killVM(serverId);
        			
        				if(resultCode.equals("REASON_0")){
        					//Kill VM successfully
        					localAttemptToTerminateVM = 0;
        				}
        				else if(resultCode.equals("REASON_395")){   
        					logger.error("Could not find the server with Id" + serverId );
        					localAttemptToTerminateVM = 0;  
        				}
        				else if(resultCode.equals("REASON_100")){   
        					logger.error(" Invalid Credentials ");
        					localAttemptToTerminateVM = 0;  
        				}
        				else if(resultCode.equals("REASON_393")){       					
        					
        					logger.error("The server " + serverId + " is associated with a Real-Server in load balancer");
        					localAttemptToTerminateVM = 0;  
        					
        				}
        				else{
        					            				
        					try{
            					
            					Thread.sleep(waitTimeToAttempt);    					
            					
            					localAttemptToTerminateVM--;
            					
            					//Failed deploy
            					logger.info("Clean Failed Deploy");
            					
            					cleanFailedVM(serverId);
            					
            				} catch (Throwable notPartOfTest ) {
            					//ignore
            				}	
        					
        				}
    					
    				}
    			
    			} catch (InternalException e) {
    				logger.error(e.getMessage());
    			} catch (CloudException e) {
    				logger.error(e.getMessage());
    				try{
    					
    					Thread.sleep(waitTimeToAttempt);    					
    					
    					localAttemptToTerminateVM--;
    					//Failed deploy
    					logger.info("Clean Failed Deploy");
    					
    					cleanFailedVM(serverId);
    					
    				} catch (Throwable notPartOfTest ) {
    					//ignore
    				}
    			} 	
    		}	
    	}    
    }
   
  private String killVM(String serverId) throws InternalException, CloudException {
       	
        HashMap<Integer, Param>  parameters = new HashMap<Integer, Param>();
        Param param = new Param(OpSource.SERVER_BASE_PATH, null);
    	parameters.put(0, param);
    	param = new Param(serverId, null);
    	parameters.put(1, param); 
    	
    	OpSourceMethod method = new OpSourceMethod(provider, 
    			provider.buildUrl(DESTROY_VIRTUAL_MACHINE,true, parameters),
    			provider.getBasicRequestParameters(OpSource.Content_Type_Value_Single_Para, "GET",null));
    	return method.requestResultCode("Terminating vm",method.invoke(),"ns9:resultCode"); 
   }
    
    private String translateZone(String zoneId) throws InternalException, CloudException {
        if( zoneId == null ) {
            for( Region r : provider.getDataCenterServices().listRegions() ) {
                zoneId = r.getProviderRegionId();
                break;
            }
        }
        if(zoneId.endsWith("a")){
        	zoneId = zoneId.substring(0, zoneId.length()-1);       	
        }
        return zoneId;
    }
    
    private VirtualMachineProduct getProduct(Architecture architecture, int cpuCout, int memoryInSize, int diskInGB) throws InternalException, CloudException{
   
        for( VirtualMachineProduct product : listProducts(architecture) ) {
            if( product.getCpuCount() == cpuCout && product.getRamInMb() == memoryInSize  && diskInGB == product.getDiskSizeInGb() ) {
                return product;
            }
        }      
       
        return null;
    }

    private VirtualMachine toVirtualMachine(Node node, Boolean isPending) throws CloudException, InternalException {
        if( node == null ) {
            return null;
        }
        HashMap<String,String> properties = new HashMap<String,String>();
        VirtualMachine server = new VirtualMachine();
        NodeList attributes = node.getChildNodes();
      
        Architecture bestArchitectureGuess = Architecture.I64;
        
        server.setTags(properties);
               
        if(isPending){
        	server.setCurrentState(VmState.PENDING);
        	server.setImagable(false);
        }else{
        	server.setCurrentState(VmState.RUNNING);
            server.setImagable(true);
        }  
        
        server.setProviderOwnerId(provider.getContext().getAccountNumber());
        server.setClonable(false);
        
        server.setPausable(true);
        server.setPersistent(true);
        
        server.setProviderRegionId(provider.getContext().getRegionId());
        
        for( int i=0; i<attributes.getLength(); i++ ) {
            Node attribute = attributes.item(i);
            if(attribute.getNodeType() == Node.TEXT_NODE) continue;
            String name = attribute.getNodeName();
            String value;
            
            if( attribute.getChildNodes().getLength() > 0 ) {
                value = attribute.getFirstChild().getNodeValue();                
            }
            else {
                continue;
            }
            //Specific server node information
            if( name.equals("id") || name.equals("id") ) {
                server.setProviderVirtualMachineId(value);                
            }
            else if( name.equalsIgnoreCase("name") ) {
                server.setName(value);
            }
            else if( name.equalsIgnoreCase("description") ) {
                server.setDescription(value);
            }
            else if( name.equalsIgnoreCase("vlanResourcePath") ) {
                
            	String vlanId = provider.getVlanIdFromVlanResourcePath(value);
            	server.setProviderVlanId(vlanId);
            
            	if(!provider.isVlanInRegion(vlanId)){
            		return null;
            	}
            	
            } 
            else if( name.equalsIgnoreCase("operatingSystem") ) {            	
           	 	NodeList osAttributes  = attribute.getChildNodes();
           	 	for(int j=0;j<osAttributes.getLength();j++ ){
           	 		Node os = osAttributes.item(j);
           	 		String osName = os.getNodeName();              
                    String osValue ;
                    if( osName.equals("displayName") && os.getChildNodes().getLength() > 0 ) {
                   	 osValue = os.getFirstChild().getNodeValue();
                    }else{
                   	 osValue = null ; 
                    }
           		 
           		 	if( osValue != null && osValue.contains("64") ) {
                      bestArchitectureGuess = Architecture.I64;
           		 	}
           		 	else if( osValue != null && osValue.contains("32") ) {
                      bestArchitectureGuess = Architecture.I32;
           		 	}
           		 	if( osValue != null ) {
           		 		server.setPlatform(Platform.guess(osValue));
           		 		break;
           		 	}           		 
           	 	}        	
            
           }
           else if( name.equalsIgnoreCase("cpuCount") ) {
        	   server.getTags().put("cpuCount", value);
           }
           else if( name.equalsIgnoreCase("memory") ) { 
        	   server.getTags().put("memory", value);
           }
           else if( name.equalsIgnoreCase("osStorage") ) { 
        	   server.getTags().put("osStorage", value);
           }
           else if( name.equalsIgnoreCase("additionalLocalStorage") ) { 
        	   server.getTags().put("additionalLocalStorage", value);
           }
           else if( name.equals("machineName") ) { 
               //equal to private ip address
           }
           else if( name.equalsIgnoreCase("privateIPAddress") ) { 
               if( value != null ) {
                   server.setPrivateIpAddresses(new String[] { value });
               }
               else {
                   server.setPrivateIpAddresses(new String[0]);
               }
               
               server.setPrivateDnsAddress(value); 
             
           }
           else if( name.equalsIgnoreCase("isDeployed") ) {
        	   
        	   if(value.equalsIgnoreCase("false")){
        		   server.setCurrentState(VmState.PENDING); 
        		   isPending = true;
        	   }else{
        		   isPending = false;
        	   }         
           }
           else if( name.equalsIgnoreCase("isStarted") ) {
        	   
        	   if(!isPending && value.equalsIgnoreCase("false")){
        		   server.setCurrentState(VmState.PAUSED);
        	   }           
           }
           else if( name.equalsIgnoreCase("created") ) {
                DateFormat df = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss'Z'"); 
                //2009-02-03T05:26:32.612278
                //2012-05-08T02:23:16.999Z
                try {
                	if(value.contains(".")){
                		String newvalue = value.substring(0,value.indexOf("."))+"Z";
                		server.setCreationTimestamp(df.parse(newvalue).getTime());                		
                	}else{
                		server.setCreationTimestamp(df.parse(value).getTime());                		
                	}                    
                }
                catch( ParseException e ) {
                    logger.warn("Invalid date: " + value);
                    server.setLastBootTimestamp(0L);
                }
           }
            //From here is the deployed server, or pending server
           else if( name.equalsIgnoreCase("machineSpecification") ) {
           		NodeList machineAttributes  = attribute.getChildNodes();
           		for(int j=0;j<machineAttributes.getLength();j++ ){
	           		Node machine = machineAttributes.item(j);
	           		
	           		if (machine.getNodeType() == Node.TEXT_NODE) continue;
	           		
	 	            if( machine.getNodeName().equalsIgnoreCase("operatingSystem") ){
	 	            	NodeList osAttributes  = machine.getChildNodes();
	 		            for(int k=0;k<osAttributes.getLength();k++ ){
	 		            	Node os = osAttributes.item(k);
	 		            	
	 		            	if(os.getNodeType() == Node.TEXT_NODE) continue;
	 		            	String osName = os.getNodeName();              
	 		                String osValue = null ;
	 		                
	 		                if( osName.equalsIgnoreCase("displayName") && os.getChildNodes().getLength() > 0 ) {
	 		                	osValue = os.getFirstChild().getNodeValue();
	 		                }else if( osName.equalsIgnoreCase("type") && os.getChildNodes().getLength() > 0) {
	 		                	osValue = os.getFirstChild().getNodeValue();
	 		                	server.setPlatform(Platform.guess(osValue));			                       
	 		                }
	 		                if( osValue != null && osValue.contains("64") ) {
	 		                	bestArchitectureGuess = Architecture.I64;
	 		                }
	 		            	else if( osValue != null && osValue.contains("32") ) {
	 		            		bestArchitectureGuess = Architecture.I32;
	 		            	}		            		     		 
	 		            }
	 	            }else if( machine.getNodeName().equalsIgnoreCase("cpuCount") && machine.getFirstChild().getNodeValue() != null ) {
	                 	server.getTags().put("cpuCount", machine.getFirstChild().getNodeValue());
	                }
	 	            //memoryMb pendingDeploy
	                else if( (machine.getNodeName().equalsIgnoreCase("memory") || machine.getNodeName().equalsIgnoreCase("memoryMb"))&& machine.getFirstChild().getNodeValue() != null ) {
	                 	server.getTags().put("memory", machine.getFirstChild().getNodeValue());
	                }
	                else if( machine.getNodeName().equalsIgnoreCase("osStorage") && machine.getFirstChild().getNodeValue() != null) { 
	              	   server.getTags().put("osStorage", machine.getFirstChild().getNodeValue());
	                }
	 	            //additionalLocalStorageGb pendingDeploy
	                else if( (machine.getNodeName().equalsIgnoreCase("additionalLocalStorage") || machine.getNodeName().equalsIgnoreCase("additionalLocalStorageGb") ) && machine.getFirstChild().getNodeValue() != null ) { 
	              	   server.getTags().put("additionalLocalStorage", machine.getFirstChild().getNodeValue());
	                }                     
           		}           
       	   }
            // pendingDeploy
           else if( name.equalsIgnoreCase("sourceImageId") ) {
        	   server.setProviderMachineImageId(value);
           }
            // pendingDeploy or Deployed
           else if( name.equalsIgnoreCase("networkId") ) {
        	   server.setProviderVlanId(value);
        	   
        	   if(!provider.isVlanInRegion(value)){
           			return null;
           		}
         
           }
          //From here is the specification for pending deployed server
           else if( name.equalsIgnoreCase("status") ) {
          		NodeList statusAttributes  = attribute.getChildNodes();
          		for(int j=0;j<statusAttributes.getLength();j++ ){
          		Node status = statusAttributes.item(j);
          		if(status.getNodeType() == Node.TEXT_NODE) continue;
	            if( status.getNodeName().equalsIgnoreCase("step") ){
	            	//TODO
	         /*   	NodeList stepAttributes  = status.getChildNodes();
		            for(int k=0;k<stepAttributes.getLength();k++ ){
		            	Node step = stepAttributes.item(j);
		            	String stepName = step.getNodeName().toLowerCase();              
		                String stepValue = null ;
		                if( stepName.equals("name") && step.getChildNodes().getLength() > 0 ) {
		                	stepValue = step.getFirstChild().getNodeValue();
		                }else if( stepName.equals("number") && step.getChildNodes().getLength() > 0) {
		                	stepValue = step.getFirstChild().getNodeValue();
		                				                       
		                }        
		            	          		     		 
		            }*/
	            }else if( status.getNodeName().equalsIgnoreCase("requestTime") && status.getFirstChild().getNodeValue() != null ) {
	            	  DateFormat df = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ssZ"); //2009-02-03T05:26:32.612278
	                  
	            	  try {
	                  	if(value.contains(".")){
	                  		String newvalue = value.substring(0,status.getFirstChild().getNodeValue().indexOf("."))+"Z";
	                  		server.setCreationTimestamp(df.parse(newvalue).getTime());                		
	                  	}else{
	                  		server.setCreationTimestamp(df.parse(status.getFirstChild().getNodeValue()).getTime());                		
	                  	}                    
	                  }
	                  catch( ParseException e ) {
	                      logger.warn("Invalid date: " + value);
	                      server.setLastBootTimestamp(0L);
	                  }
	            }  
	            else if( status.getNodeName().equalsIgnoreCase("userName") && status.getFirstChild().getNodeValue() != null ) {
                	server.setProviderOwnerId(status.getFirstChild().getNodeValue());
	            }
	            else if( status.getNodeName().equalsIgnoreCase("numberOfSteps") ) { 
	             	
	            }
	            else if( status.getNodeName().equalsIgnoreCase("action") ) { 
	            	String action = status.getFirstChild().getNodeValue();
	            	if(action.equalsIgnoreCase("CLEAN_SERVER")){
	            		//Means failed deployed
	            		server.setCurrentState(VmState.PENDING);	   
	            	}
                }
                                 
          	  }           
      	   }
        }      
        
        if( server.getName() == null ) {
            server.setName(server.getProviderVirtualMachineId());
        }
        if( server.getDescription() == null ) {
            server.setDescription(server.getName());
        }
        server.setProviderAssignedIpAddressId(null);
       
        if( server.getProviderDataCenterId() == null ) {
            server.setProviderDataCenterId(provider.getDataCenterId(server.getProviderRegionId()));
        }        
        
        if( server.getPlatform() == null && server.getName() != null ) {
        	server.setPlatform(Platform.guess(server.getName()));        	
        }
        else {
        	server.setPlatform(Platform.UNKNOWN);
        }
        if( server.getArchitecture() == null ) {
        	server.setArchitecture(bestArchitectureGuess);
        }
        
        
        VirtualMachineProduct product = new VirtualMachineProduct();
        if(server.getTag("cpuCount") != null && server.getTag("memory") != null ){
        	int cpuCout = Integer.valueOf((String) server.getTag("cpuCount"));
    		int memoryInMb = Integer.valueOf((String) server.getTag("memory"));
        	
        	if(server.getTag("additionalLocalStorage") == null){
        		product = getProduct(bestArchitectureGuess, cpuCout, memoryInMb, 0);
        		
        	}else{
        		int diskInGb = Integer.valueOf((String) server.getTag("additionalLocalStorage"));
        		product = getProduct(bestArchitectureGuess, cpuCout, memoryInMb, diskInGb); 
        	}
        }
        
        server.setProduct(product);        
        
        return server;
    }       
  
}
